from hierarchybuilder.combine_spans import utils as combine_spans_utils
from hierarchybuilder import utils as ut
import hierarchybuilder.DAG.NounPhraseObject as NounPhrase
from hierarchybuilder.visualization import normalize_quantity
import torch


def add_node_between_nodes(parent, child, new_node):
    parent.children.remove(child)
    child.parents.remove(parent)
    parent.children.append(new_node)
    child.parents.add(new_node)
    new_node.children.append(child)
    new_node.parents.add(parent)
    new_node.label_lst.update(child.label_lst)


def is_node_contained_in_another_node(node, ref_node):
    if combine_spans_utils.is_similar_meaning_between_span(node.common_lemmas_in_spans,
                                                           ref_node.common_lemmas_in_spans):
        return True
    return False


def locate_node_in_DAG(parent, new_node):
    update_lst = set()
    for child in parent.children:
        if is_node_contained_in_another_node(new_node, child):
            update_lst.add(child)
    if update_lst:
        for node in update_lst:
            add_node_between_nodes(parent, node, new_node)
        return
    parent.add_children([new_node])
    new_node.parents.add(parent)


def add_NP_to_DAG_bottom_to_up(np_object_to_add, np_object, visited, similar_np_object, visited_and_added=set()):
    is_contained = False
    if np_object == np_object_to_add:
        return True
    if np_object in visited_and_added:
        return True
    if np_object in visited:
        return False
    visited.add(np_object)
    if not np_object_to_add.common_lemmas_in_spans:
        # print("Empty node")
        # print(np_object_to_add.span_lst)
        # print(np_object_to_add.lemma_to_occurrences_dict)
        return False
    if combine_spans_utils.is_similar_meaning_between_span(np_object_to_add.common_lemmas_in_spans,
                                                           np_object.common_lemmas_in_spans):
        is_contained = True
        if len(np_object_to_add.common_lemmas_in_spans) == len(np_object.common_lemmas_in_spans):
            np_object.combine_nodes(np_object_to_add)
            similar_np_object[0] = np_object
            return True
    if is_contained:
        is_added = False
        for parent in np_object.parents:
            is_added |= add_NP_to_DAG_bottom_to_up(np_object_to_add, parent, visited, similar_np_object,
                                                   visited_and_added)
            if similar_np_object[0]:
                return True
        if not is_added:
            np_object_to_add.add_children([np_object])
            parents_remove_lst = set()
            for parent in np_object.parents:
                if parent in np_object_to_add.parents:
                    parents_remove_lst.add(parent)
                    continue
                if is_node_contained_in_another_node(parent, np_object_to_add):
                    np_object_to_add.parents.add(parent)
                    if np_object_to_add not in parent.children:
                        parent.children.append(np_object_to_add)
                    else:
                        continue
                    parents_remove_lst.add(parent)
            for parent_object in parents_remove_lst:
                np_object.parents.remove(parent_object)
                if np_object in parent_object.children:
                    parent_object.children.remove(np_object)
            np_object.parents.add(np_object_to_add)
        visited_and_added.add(np_object)
        return True
    return False


def create_np_object_from_np_collection(np_collection, dict_span_to_lemmas_lst, labels, span_to_object):
    tuple_np_lst = []
    for np in np_collection:
        tuple_np_lst.append((np, dict_span_to_lemmas_lst[np]))
    np_object = NounPhrase.NP(tuple_np_lst, labels)
    for np in np_collection:
        span_to_object[np] = np_object
    return np_object


def update_np_object(collection_np_object, np_collection, span_to_object, dict_span_to_lemmas_lst, labels):
    collection_np_object.label_lst.update(labels)
    nps_to_update = set()
    for np in np_collection:
        np_object = span_to_object.get(np, None)
        if np_object == collection_np_object:
            continue
        if np_object:
            np_object.combine_nodes(collection_np_object)
            for span in np_object.span_lst:
                span_to_object[span] = np_object
            collection_np_object = np_object
        nps_to_update.add(np)
    for np in nps_to_update:
        span_to_object[np] = collection_np_object
        collection_np_object.span_lst.add(np)
        collection_np_object.add_unique_lst([dict_span_to_lemmas_lst[np]])


def update_global_label_with_its_object(global_dict_label_to_object, np_object, dict_object_to_global_label,
                                        longest_np_lst, longest_NP_to_global_index):
    uncounted_labels = set()
    for label in np_object.label_lst:
        node = global_dict_label_to_object.get(label, None)
        if node:
            continue
        uncounted_labels.add(label)
    global_indices_in_np_object = set()
    for span in np_object.span_lst:
        if span in longest_np_lst:
            label = longest_NP_to_global_index.get(span, -1)
            if label == -1:
                continue
            global_indices_in_np_object.add(label)
    uncounted_global_labels = uncounted_labels.intersection(global_indices_in_np_object)
    for label in uncounted_global_labels:
        global_dict_label_to_object[label] = np_object

    dict_object_to_global_label[hash(np_object)] = uncounted_global_labels


def create_and_insert_nodes_from_sub_groups_of_spans(dict_score_to_collection_of_sub_groups,
                                                     dict_span_to_lemmas_lst, span_to_object,
                                                     dict_span_to_similar_spans, global_dict_label_to_object,
                                                     dict_object_to_global_label, longest_np_lst,
                                                     objects_longest_nps_from_previous_clusters,
                                                     longest_NP_to_global_index):
    longest_nps_total_lst = objects_longest_nps_from_previous_clusters.copy()
    np_objet_lst = []
    for score, np_to_labels_collection in dict_score_to_collection_of_sub_groups.items():
        for np_key, labels in np_to_labels_collection:
            np_collection = dict_span_to_similar_spans[np_key]
            np_object = None
            for np in np_collection:
                np_object = span_to_object.get(np, None)
                if np_object:
                    update_np_object(np_object, np_collection, span_to_object, dict_span_to_lemmas_lst, labels)
                    break
            if np_object:
                continue
            np_object = create_np_object_from_np_collection(np_collection, dict_span_to_lemmas_lst, labels,
                                                            span_to_object)
            np_objet_lst.append(np_object)
    np_objet_lst = sorted(np_objet_lst, key=lambda np_object: np_object.length, reverse=True)
    for np_object in np_objet_lst:
        update_global_label_with_its_object(global_dict_label_to_object, np_object, dict_object_to_global_label,
                                            longest_np_lst, longest_NP_to_global_index)
        similar_np_object = [None]
        visited = set()
        visited_and_added = set()
        np_object_temp = np_object
        for longest_nps_node in longest_nps_total_lst:
            if longest_nps_node in visited:
                continue
            add_NP_to_DAG_bottom_to_up(np_object_temp, longest_nps_node, visited, similar_np_object, visited_and_added)
            if similar_np_object[0]:
                # for span in np_object_temp.span_lst:
                #     span_to_object[span] = similar_np_object[0]
                # if hash(np_object_temp) in dict_object_to_global_label:
                #     for label in dict_object_to_global_label[hash(np_object_temp)]:
                #         global_dict_label_to_object[label] = similar_np_object[0]
                #     dict_object_to_global_label[hash(similar_np_object[0])] = \
                #         dict_object_to_global_label.get(hash(similar_np_object[0]), set())
                #     dict_object_to_global_label[hash(similar_np_object[0])].update(
                #         dict_object_to_global_label[hash(np_object_temp)])
                #     dict_object_to_global_label.pop(hash(np_object_temp), None)
                update_global_dictionaries(span_to_object, global_dict_label_to_object,
                                           dict_object_to_global_label, np_object_temp, similar_np_object[0])
                if similar_np_object[0]:
                    np_object_temp = similar_np_object[0]
                    break
        if np_object_temp not in longest_nps_total_lst and np_object_temp.span_lst.intersection(longest_np_lst):
            longest_nps_total_lst.add(np_object)


def remove_topic_np_from_np_object(np_object, topic_np):
    lemmas_to_remove = None
    for span_as_lemmas_lst in np_object.list_of_span_as_lemmas_lst:
        if len(span_as_lemmas_lst) == 1 and topic_np in span_as_lemmas_lst:
            lemmas_to_remove = span_as_lemmas_lst
            break
    if lemmas_to_remove:
        np_object.list_of_span_as_lemmas_lst.remove(lemmas_to_remove)


def update_global_dictionaries(span_to_object, global_dict_label_to_object, dict_object_to_global_label,
                               previous_np_object, new_np_object):
    for span in previous_np_object.span_lst:
        span_to_object[span] = new_np_object
    if hash(previous_np_object) in dict_object_to_global_label:
        for label in dict_object_to_global_label[hash(previous_np_object)]:
            global_dict_label_to_object[label] = new_np_object
        dict_object_to_global_label[hash(new_np_object)] = \
            dict_object_to_global_label.get(hash(new_np_object), set())
        dict_object_to_global_label[hash(new_np_object)].update(
            dict_object_to_global_label[hash(previous_np_object)])
        dict_object_to_global_label.pop(hash(previous_np_object), None)


def combine_topic_object(np_object, topic_object):
    if combine_spans_utils.is_similar_meaning_between_span(topic_object.common_lemmas_in_spans,
                                                           np_object.common_lemmas_in_spans):
        if len(topic_object.common_lemmas_in_spans) == len(np_object.common_lemmas_in_spans):
            np_object.combine_nodes(topic_object)
            return True
    return False


def create_and_update_topic_object(topic_synonym_lst, span_to_object, longest_NP_to_global_index,
                                   global_dict_label_to_object, dict_object_to_global_label):
    topic_synonyms_tuples = [(synonym, [synonym]) for synonym in topic_synonym_lst]
    topic_object = NounPhrase.NP(topic_synonyms_tuples, set())
    span_lst = topic_object.span_lst.copy()
    for np in span_lst:
        np_object = span_to_object.get(np, None)
        if np_object:
            if np_object == topic_object:
                continue
            is_combined = combine_topic_object(np_object, topic_object)
            if is_combined:
                update_global_dictionaries(span_to_object, global_dict_label_to_object, dict_object_to_global_label,
                                           topic_object, np_object)
                topic_object = np_object
            else:
                remove_topic_np_from_np_object(np_object, np)
    for np in topic_object.span_lst:
        span_to_object[np] = topic_object
        label = longest_NP_to_global_index.get(np, None)
        if label:
            topic_object.label_lst.add(label)
    return topic_object


def insert_examples_of_topic_to_DAG(dict_score_to_collection_of_sub_groups, topic_synonym_lst, dict_span_to_lemmas_lst,
                                    span_to_object, dict_span_to_similar_spans,
                                    global_dict_label_to_object, topic_object_lst,
                                    longest_np_total_lst, longest_np_lst, longest_NP_to_global_index,
                                    dict_object_to_global_label):
    longest_nps_from_previous_clusters = set(longest_np_total_lst) - set(longest_np_lst)
    objects_longest_nps_from_previous_clusters = set()
    for longest_np in longest_nps_from_previous_clusters:
        objects_longest_nps_from_previous_clusters.add(span_to_object[longest_np])
    create_and_insert_nodes_from_sub_groups_of_spans(dict_score_to_collection_of_sub_groups,
                                                     dict_span_to_lemmas_lst, span_to_object,
                                                     dict_span_to_similar_spans,
                                                     global_dict_label_to_object, dict_object_to_global_label,
                                                     longest_np_lst, objects_longest_nps_from_previous_clusters,
                                                     longest_NP_to_global_index)
    topic_object = create_and_update_topic_object(topic_synonym_lst, span_to_object, longest_NP_to_global_index,
                                                  global_dict_label_to_object, dict_object_to_global_label)
    add_dependency_routh_between_longest_np_to_topic(span_to_object, topic_object_lst,
                                                     longest_np_total_lst, topic_object, dict_object_to_global_label,
                                                     global_dict_label_to_object)


def add_descendants_of_node_to_graph(node, global_index_to_similar_longest_np, new_taxonomic_np_objects,
                                     different_concepts, concept_to_occurrences):
    node.span_lst = list(node.span_lst)
    node.span_lst.sort(key=lambda span_lst: ut.dict_span_to_counter.get(span_lst, 0), reverse=True)
    span_to_present = normalize_quantity.normalized_quantity_node(node)
    for span in node.span_lst:
        if span == span_to_present:
            continue
        span_to_present += " | "
        span_to_present += span
    label_lst = get_labels_of_children(node.children)
    label_lst = node.label_lst - label_lst
    NP_occurrences = get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                   label_lst)
    covered_occurrences = get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                        node.label_lst)
    if NP_occurrences:
        if node not in different_concepts:
            concept_to_occurrences[span_to_present] = NP_occurrences
        different_concepts.add(node)
    if covered_occurrences != NP_occurrences:
        span_to_present += " (" + str(NP_occurrences) + "/ " + str(covered_occurrences) + ")"
    else:
        span_to_present += " (" + str(NP_occurrences) + ")"
    np_val_dict = {span_to_present: {}}
    node.children = sorted(node.children, key=lambda child: get_frequency_from_labels_lst(
        global_index_to_similar_longest_np,
        child.label_lst), reverse=True)
    for child in node.children:
        np_val_dict[span_to_present].update(add_descendants_of_node_to_graph(child, global_index_to_similar_longest_np,
                                                                             new_taxonomic_np_objects,
                                                                             different_concepts,
                                                                             concept_to_occurrences))
    return np_val_dict


def from_DAG_to_JSON(topic_object_lst, global_index_to_similar_longest_np, new_taxonomic_np_objects, different_concepts,
                     concept_to_occurrences):
    np_val_lst = {}
    topic_object_lst.sort(key=lambda topic_object: topic_object.marginal_val, reverse=True)
    for topic_node in topic_object_lst:
        np_val_lst.update(add_descendants_of_node_to_graph(topic_node, global_index_to_similar_longest_np,
                                                           new_taxonomic_np_objects, different_concepts,
                                                           concept_to_occurrences))
    return np_val_lst


def add_dependency_routh_between_longest_np_to_topic(span_to_object, topic_object_lst,
                                                     longest_nps, topic_object, dict_object_to_global_label,
                                                     global_dict_label_to_object):
    visited_and_added = set()
    visited = set()
    for longest_np_span in longest_nps:
        np_object = span_to_object[longest_np_span]
        if np_object in visited:
            continue
        if np_object in topic_object_lst:
            continue
        similar_np_object = [None]
        add_NP_to_DAG_bottom_to_up(topic_object, np_object, visited, similar_np_object, visited_and_added)
        if similar_np_object[0]:
            update_global_dictionaries(span_to_object, global_dict_label_to_object, dict_object_to_global_label,
                                       topic_object, similar_np_object[0])
            if topic_object in topic_object_lst:
                topic_object_lst.remove(topic_object)
            topic_object = similar_np_object[0]
    for np in topic_object.span_lst:
        span_to_object[np] = topic_object
    if topic_object not in topic_object_lst:
        topic_object_lst.append(topic_object)


def update_score(topic_object_lst, dict_span_to_rank, visited=[]):
    for node in topic_object_lst:
        if node in visited:
            continue
        visited.append(node)
        node.score = combine_spans_utils.get_average_value(node.span_lst, dict_span_to_rank)
        update_score(node.children, dict_span_to_rank, visited)


def check_symmetric_relation_in_DAG(nodes_lst, visited=set()):
    for node in nodes_lst:
        if node in visited:
            continue
        if node in node.parents:
            print(node.span_lst)
            raise Exception("node can't be itself parent")
        if node in node.children:
            raise Exception("node can't be itself child")
        visited.add(node)
        for child in node.children:
            if node not in child.parents:
                print("child spans:")
                print(child.span_lst)
                print("node spans:")
                print(node.span_lst)
                raise Exception("Parent and child isn't consistent")
        for parent in node.parents:
            if node not in parent.children:
                print(parent.span_lst)
                print(node.span_lst)
                raise Exception("Parent and child isn't consistent")
        check_symmetric_relation_in_DAG(node.children, visited)


def update_symmetric_relation_in_DAG(nodes_lst, visited=set()):
    for node in nodes_lst:
        if node in visited:
            continue
        visited.add(node)
        if node in node.parents:
            node.parents.remove(node)
            print("node is found in its parents' list")
            print(node.span_lst)
        if node in node.children:
            while node not in node.children:
                node.children.remove(node)
            print("node is found in its children's list")
            print(node.span_lst)
        for child in node.children:
            if node not in child.parents:
                child.parents.add(node)
                print("node isn't found in its children parents' list")
                print("Node:")
                print(node.span_lst)
                print("Child:")
                print(child.span_lst)
        for parent in node.parents:
            if node not in parent.children:
                parent.children.append(node)
                print("parent isn't found in its parent children's list")
                print(node.span_lst)
        update_symmetric_relation_in_DAG(node.children, visited)


def get_frequency_from_labels_lst(global_index_to_similar_longest_np, label_lst):
    num_of_labels = 0
    for label in label_lst:
        for span in global_index_to_similar_longest_np.get(label, []):
            try:
                num_of_labels += ut.dict_longest_span_to_counter.get(span, 1)
            except:
                print(span)
                raise Exception("the error is found in the get counter of longest span")
    return num_of_labels


def initialize_node_weighted_vector(node, span_to_vector):
    if not node.span_lst:
        return
    is_first = True
    weighted_average_vector = torch.zeros(768, 1)
    for np in node.span_lst:
        if is_first:
            is_first = False
            span_vector = span_to_vector.get(np, None)
            if span_vector is None:
                # print("the following span is none")
                # print(np)
                # weighted_average_vector = torch.zeros(768, 1)
                continue
            weighted_average_vector = span_vector
        else:
            span_vector = span_to_vector.get(np, None)
            if span_vector is None:
                # print("the following span is none")
                # print(np)
                continue
            weighted_average_vector += span_vector
    weighted_average_vector /= len(node.span_lst)
    node.weighted_average_vector = weighted_average_vector


def initialize_all_spans_vectors(all_spans, span_to_vector):
    with torch.no_grad():
        for i in range(0, len(all_spans), 100):
            batch = all_spans[i: i + 100]
            encoded_input = \
                ut.sapBert_tokenizer(batch, return_tensors='pt', padding=True).to(ut.device)
            temp = ut.model(**encoded_input)
            phrase_embeddings = temp.last_hidden_state.cpu()
            phrase_embeddings = torch.transpose(phrase_embeddings, 0, 1)
            phrase_embeddings = phrase_embeddings[0, :]
            del encoded_input, temp
            torch.cuda.empty_cache()
            for j in range(len(batch)):
                span_to_vector[batch[j]] = phrase_embeddings[j].reshape(-1, 1)


def get_represented_vector(span):
    with torch.no_grad():
        encoded_input = \
            ut.sapBert_tokenizer(span, return_tensors='pt', padding=True).to(ut.device)
        temp = ut.model(**encoded_input)
        phrase_embeddings = temp.last_hidden_state.cpu()
        phrase_embeddings = torch.transpose(phrase_embeddings, 0, 1)
        phrase_embeddings = phrase_embeddings[0, :]
        del encoded_input, temp
        torch.cuda.empty_cache()
    return phrase_embeddings.reshape(-1, 1)


def initialize_nodes_weighted_average_vector(topic_object_lst, global_index_to_similar_longest_np, span_to_vector,
                                             visited=[]):
    for node in topic_object_lst:
        if node in visited:
            continue
        initialize_node_weighted_vector(node, span_to_vector)
        node.frequency = get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                       node.label_lst)
        visited.append(node)
        initialize_nodes_weighted_average_vector(node.children, global_index_to_similar_longest_np, span_to_vector,
                                                 visited)


def get_labels_of_children(children):
    label_lst = set()
    for child in children:
        label_lst.update(child.label_lst)
    return label_lst


def get_labels_of_leaves(children):
    label_lst = set()
    for child in children:
        if not child.children:
            label_lst.update(child.label_lst)
    return label_lst


def get_leaves_from_DAG(nodes_lst, leaves_lst=set(), visited=set()):  # function for dfs
    for node in nodes_lst:
        if node not in visited:
            if len(node.children) == 0:
                leaves_lst.add(node)
            visited.add(node)
        else:
            continue
        get_leaves_from_DAG(node.children, leaves_lst, visited)
    return leaves_lst


def remove_redundant_nodes(nodes_lst):
    leaves_lst = get_leaves_from_DAG(nodes_lst)
    queue = []
    visited = []
    visited.extend(leaves_lst)
    queue.extend(leaves_lst)
    counter = 0
    ignore_lst = []
    while queue:
        counter += 1
        s = queue.pop(0)
        parents = s.parents.copy()
        for parent in parents:
            if parent in ignore_lst:
                continue
            if len(set(parent.children)) == 1 and parent.label_lst == s.label_lst:
                if parent.parents:
                    for ancestor in parent.parents:
                        ancestor.add_children([s])
                        if parent in ancestor.children:
                            ancestor.children.remove(parent)
                        s.parents.add(ancestor)
                    remove_lst = parent.parents.copy()
                    for ancestor in remove_lst:
                        if ancestor in parent.parents:
                            parent.parents.remove(ancestor)
                    if parent in nodes_lst:
                        nodes_lst.remove(parent)
                    ignore_lst.append(parent)
                    if parent in s.parents:
                        s.parents.remove(parent)
                    if s in parent.children:
                        parent.children.remove(s)
                    queue.append(s)
                else:
                    ignore_lst.append(parent)
                    if parent in s.parents:
                        s.parents.remove(parent)
                    if parent in nodes_lst:
                        nodes_lst.remove(parent)
                    if not s.parents:
                        nodes_lst.append(s)
                        break
            else:
                if parent not in visited:
                    visited.append(parent)
                    queue.append(parent)



import argparse
import pandas as pd
from tqdm import tqdm

dict_id_to_similar_concepts = {}
dict_concept_to_id = {}
dict_id_to_concept = {}
dict_from_relation_type_between_terms = {}
dict_explicit_relation_type_between_terms = {}
umls_loader = None


def parse_args(umls_relation_file_path, umls_synonymous_file_path):
    parser = argparse.ArgumentParser()
    parser.add_argument('--umls_dict_dir', type=str,
                        default=umls_synonymous_file_path,
                        help="the directory of the umls dictionary corpus")
    parser.add_argument('--umls_relation_dir', type=str,
                        default=umls_relation_file_path,
                        help="the directory of the umls relation corpus")
    return parser.parse_args()


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class UMLSLoader(metaclass=Singleton):
    def __init__(self, corpus_dir_dict, corpus_dir_relation):
        tqdm.pandas()
        self.umls_df_chunks: pd.DataFrame = self.read_df_for_dict(corpus_dir_dict)
        self.create_dicts_from_pd_table()
        self.read_df_for_relation(corpus_dir_relation)

    @staticmethod
    def read_df_for_dict(corpus_dir):
        print("start read dict table")
        col_names = ['CUI', 'LAT', 'TS', 'LUI', 'STT', 'SUI', 'ISPREF', 'AUI', 'SAUI', 'SCUI', 'SDUI', 'SAB', 'TTY',
                     'CODE', 'STR', 'SRL', 'SUPPRESS', 'CVF']
        umls_df_chunks = pd.read_table(corpus_dir, sep='|', chunksize=1000000, encoding='utf-8', index_col=False,
                                       header=None, names=col_names, low_memory=False)
        print("End read dict table")
        return umls_df_chunks

    @staticmethod
    def read_df_for_relation(corpus_dir):
        print("Start read table for relation")
        col_names = ['CUI1', 'AUI1', 'STYPE1', 'REL', 'CUI2', 'AUI2', 'STYPE2', 'RELA', 'RUI', 'SRUI', 'SAB', 'SL',
                     'RG', 'DIR', 'SUPPRESS', 'CVF']
        chunks = pd.read_table(corpus_dir, sep='|', chunksize=1000000, encoding='utf-8', index_col=False, header=None,
                               names=col_names,
                               low_memory=False)
        counter = 0
        num_of_relation = 0
        all_relations = set()
        for chunk in chunks:
            indexed_df = chunk.reset_index()
            for index, row in indexed_df.iterrows():
                relation_type = row['REL']
                all_relations.add(relation_type)
                dict_from_relation_type_between_terms[relation_type] = dict_from_relation_type_between_terms.get(
                    relation_type, {})
                first_concept_cui = row['CUI1']
                num_of_relation += 1
                dict_from_relation_type_between_terms[relation_type][first_concept_cui] = \
                    dict_from_relation_type_between_terms[relation_type].get(first_concept_cui, set())
                dict_from_relation_type_between_terms[relation_type][first_concept_cui].add(row['CUI2'])
            counter += 1
            print(counter)
        print(num_of_relation)
        print(all_relations)

    def create_dicts_from_pd_table(self):
        counter = 0
        ENG_is_valid = False
        for chunk in self.umls_df_chunks:
            chunk_temp = chunk
            chunk_temp['STR'] = chunk_temp['STR'].str.lower()
            indexed_df = chunk_temp.reset_index()
            for index, row in indexed_df.iterrows():
                if row['LAT'] != 'ENG':
                    continue
                if not ENG_is_valid:
                    # print("ENG is valid ENTRY")
                    ENG_is_valid = True
                dict_id_to_similar_concepts[row['CUI']] = dict_id_to_similar_concepts.get(row['CUI'], set())
                dict_id_to_similar_concepts[row['CUI']].add(row['STR'])
                dict_concept_to_id[row['STR']] = row['CUI']
            counter += 1
            print(counter)

    def get_term_aliases(self, term):
        cuis = dict_concept_to_id.get(term, None)
        if not cuis:
            return set()
        else:
            return dict_id_to_similar_concepts[cuis]

    def get_broader_term(self, term, relation_type):
        if dict_from_relation_type_between_terms.get(relation_type, None):
            relation_type_dict = dict_from_relation_type_between_terms[relation_type]
            term_cui = dict_concept_to_id.get(term, None)
            if not term_cui:
                return set()
            related_term_cui_lst = relation_type_dict.get(term_cui, None)
            if related_term_cui_lst:
                broader_terms_lst = []
                for related_term_cui in related_term_cui_lst:
                    similar_concepts = dict_id_to_similar_concepts.get(related_term_cui, None)
                    if similar_concepts:
                        broader_terms_lst.append(similar_concepts)
                if broader_terms_lst:
                    return broader_terms_lst
        return set()


def load_umls_data(umls_relations_file_path, umls_synonymous_file_path):
    global umls_loader
    args = parse_args(umls_relations_file_path, umls_synonymous_file_path)
    umls_loader = UMLSLoader(args.umls_dict_dir, args.umls_relation_dir)



from hierarchybuilder.expansions import valid_expansion_utils
from nltk.corpus import stopwords

stop_words = set(stopwords.words('english'))
tied_deps = ['compound', 'mwe', 'case', 'mark', 'auxpass', 'name', 'aux', 'neg', 'det', 'goeswith', 'nummod',
             'quantmod', 'nmod:npmod']
tied_couples = [['auxpass', 'nsubjpass']]

dep_type_optional = ['advmod', 'dobj', 'npadvmod', 'nmod', 'conj', 'aux', 'poss', 'nmod:poss',
                     'xcomp']  # , 'conj', 'nsubj', 'appos'

acl_to_seq = ['acomp', 'dobj', 'nmod']  # acl and relcl + [[['xcomp'], ['aux']], 'dobj']
dep_to_seq = ['quantmod', 'cop']  # 'cc',
combined_with = ['acl', 'relcl', 'acl:relcl', 'ccomp', 'advcl', 'amod']  # +, 'cc'
couple_to_seq = {'quantmod': ['amod'], 'cop': ['nsubjpass']}  # 'nsubjpass': ['amod'] ,'cc': ['conj', 'nmod']

low_val_dep = ['neg', 'nmod:poss', 'case', 'mark', 'auxpass', 'aux', 'nummod', 'quantmod', 'cop']
med_val_dep = ['nsubjpass', 'advmod', 'npadvmod', 'conj', 'poss', 'nmod:poss', 'xcomp', 'nmod:npmod', 'dobj', 'nmod',
               'amod', 'nsubj', 'acl', 'relcl', 'acl:relcl', 'ccomp', 'advcl']
max_val_dep = ['compound', 'mwe', 'name']
neglect_deps = ['neg', 'case', 'mark', 'auxpass', 'aux', 'cop']


def get_tied_couple_by_deps(first_dep, second_dep, children):
    first_token = None
    second_token = None
    for child in children:
        if child.dep_ == first_dep:
            first_token = child
        if child.dep_ == second_dep:
            second_token = child
    return first_token, second_token


def get_tied_couples(children):
    tied_couples_to_add = []
    for dep_couples in tied_couples:
        first_token, second_token = get_tied_couple_by_deps(dep_couples[0], dep_couples[1], children)
        if not first_token or not second_token:
            continue
        tied_couples_to_add.append(first_token)
        tied_couples_to_add.append(second_token)
    return tied_couples_to_add


tied_compound = ['compound', 'mwe', 'name']


def is_dep_tied_by_tied_preposition(token):
    for child in token.children:
        if child.dep_ in ['case', 'mark']:
            if child.text.lower() == 'of':
                return True
        if child.dep_ in tied_compound:
            is_dep_tied = is_dep_tied_by_tied_preposition(child)
            if is_dep_tied:
                return True
    return False


def combine_tied_deps_recursively_and_combine_their_children(head, boundary_np_to_the_left):
    combined_children_lst = []
    combined_tied_tokens = [head]
    tied_couples_to_add = get_tied_couples(head.children)
    for child in head.children:
        if boundary_np_to_the_left > child.i:
            continue
        if child.dep_ in tied_deps or child in tied_couples_to_add:
            temp_tokens, temp_children = combine_tied_deps_recursively_and_combine_their_children(
                child, boundary_np_to_the_left)
            combined_tied_tokens.extend(temp_tokens)
            combined_children_lst.extend(temp_children)
        else:
            if is_dep_tied_by_tied_preposition(child):
                temp_tokens, temp_children = combine_tied_deps_recursively_and_combine_their_children(
                    child, boundary_np_to_the_left)
                combined_tied_tokens.extend(temp_tokens)
                combined_children_lst.extend(temp_children)
            else:
                combined_children_lst.append(child)
    return combined_tied_tokens, combined_children_lst


def get_if_has_preposition_child_between(token, head):
    prep_lst = []
    for child in token.children:
        if child.dep_ == 'case' and child.tag_ == 'IN':
            prep_lst.append(child)
    if prep_lst:
        first_prep_child_index = 100
        prep_child = None
        for child in prep_lst:
            if child.i < first_prep_child_index:
                prep_child = child
        if token.i > head.i:
            first_val = head.i
            second_val = token.i
        else:
            first_val = token.i
            second_val = head.i
        if first_val < prep_child.i < second_val:
            return prep_child
        else:
            return None
    return None


def initialize_couple_lst(others, couple_lst, lst_children):
    for other in others:
        dep_type = couple_to_seq[other.dep_]
        for token in lst_children:
            if token.dep_ in dep_type:
                couple_lst.append([other, token])


def get_couple_seq_if_exist(lst_children):
    dep_to_seq_lst = []
    for child in lst_children:
        if child.dep_ in dep_to_seq:
            dep_to_seq_lst.append(child)
    remove_lst = []
    couple_seq_lst = []
    for dep_child in dep_to_seq_lst:
        dep_type = couple_to_seq[dep_child.dep_]
        remove_lst_temp = []
        for token in lst_children:
            if token.dep_ in dep_type:
                couple_seq_lst.append([dep_child, token])
                remove_lst_temp = [dep_child, token]
                break
        if remove_lst_temp:
            for token in remove_lst_temp:
                lst_children.remove(token)
            remove_lst.extend(remove_lst_temp)
    return couple_seq_lst


def get_appos_if_exist(lst_children):
    appos_child_lst = []
    for child in lst_children:
        if child.dep_ == 'appos':
            appos_child_lst.append(child)
    if appos_child_lst:
        for child in lst_children:
            if child.dep_ == 'dep':
                appos_child_lst.append(child)
        lst_children_copy = lst_children.copy()
        for child in lst_children_copy:
            if child in appos_child_lst:
                lst_children.remove(child)
        if len(appos_child_lst) == 1:
            if appos_child_lst[0].pos_ == 'PROPN':
                return []
    return appos_child_lst


def get_conj_if_cc_exist(lst_children):
    cc_is_exist = False
    cc_child_lst = []
    for child in lst_children:
        if child.dep_ == 'cc' or child.dep_ == 'punct' and child.text == ',':
            cc_child_lst.append(child)
            cc_is_exist = True
    if cc_is_exist:
        children_dep = valid_expansion_utils.get_token_by_dep(lst_children, 'conj')
        if children_dep is []:
            children_dep = valid_expansion_utils.get_token_by_dep(lst_children, 'nmod')
        children_dep.sort(key=lambda x: x.i)
        cc_child_lst.sort(key=lambda x: x.i)
        tokens_to_skip = cc_child_lst.copy()
        conj_tokens_lst = []
        for cc_child in cc_child_lst:
            for child in children_dep:
                if child.i > cc_child.i:
                    tokens_to_skip.append(child)
                    conj_tokens_lst.append([cc_child, child])
                    children_dep.remove(child)
                    break
        for token in tokens_to_skip:
            lst_children.remove(token)
        return conj_tokens_lst
    return []


def get_seq_deps(lst_children):
    seq_dict = {}
    conj_tokens_lst = get_conj_if_cc_exist(lst_children)
    if conj_tokens_lst:
        seq_dict['conj'] = conj_tokens_lst
    couple_seq_lst = get_couple_seq_if_exist(lst_children)
    if couple_seq_lst:
        seq_dict['couple_seq'] = couple_seq_lst
    appos_child_lst = get_appos_if_exist(lst_children)
    if appos_child_lst:
        seq_dict['appos'] = appos_child_lst
    return seq_dict


def get_all_dominated_deps_from_lst(token, deps_collection):
    deps_collection.append(token)
    for child_token in token.children:
        get_all_dominated_deps_from_lst(child_token, deps_collection)


def set_couple_deps(seq_dict, boundary_np_to_the_left, sub_np_lst, head):
    for seq_type, seq_lst in seq_dict.items():
        if seq_type == 'appos':
            deps_collection = []
            for token in seq_lst:
                get_all_dominated_deps_from_lst(token, deps_collection)
            sub_np_lst.append(deps_collection)
            continue
        for seq in seq_lst:
            sub_np_lst_seq = []
            all_sub_of_sub = []
            for token in seq:
                sub_np_lst_token, lst_children = \
                    combine_tied_deps_recursively_and_combine_their_children(token, boundary_np_to_the_left)
                sub_np_lst_seq.extend(sub_np_lst_token)
                get_children_expansion(all_sub_of_sub, lst_children, boundary_np_to_the_left, head)
            if all_sub_of_sub:
                sub_np_lst_seq.append(all_sub_of_sub)
            sub_np_lst.append(sub_np_lst_seq)


def get_all_valid_sub_special(token, boundary_np_to_the_left):
    sub_np_lst, lst_children = combine_tied_deps_recursively_and_combine_their_children(token, boundary_np_to_the_left)
    complete_children = []
    seq_dict = get_seq_deps(lst_children)
    complete_occurrences = 0
    for child in lst_children:
        if child.text in stop_words:
            continue
        if child.dep_ in ['dobj', 'advcl', 'nmod']:  # 'cc', 'conj', 'aux', 'auxpass', 'cop', 'nsubjpass'
            all_sub_of_sub = get_all_valid_sub_np(child, boundary_np_to_the_left)
            if complete_occurrences > 0 and child.dep_ in ['nmod', 'advcl'] and len(all_sub_of_sub) > 1:
                new_sub_np_lst = []
                optional_lst = []
                for item in all_sub_of_sub:
                    if isinstance(item, list):
                        optional_lst.append(item)
                    else:
                        new_sub_np_lst.append(item)
                new_sub_np_lst.sort(key=lambda x: x.i)
                new_sub_np_lst = new_sub_np_lst + optional_lst
                if new_sub_np_lst[0].dep_ in ['case', 'mark']:
                    sub_np_lst = sub_np_lst + [new_sub_np_lst]
                    complete_occurrences += 1
                    continue
            sub_np_lst = sub_np_lst + all_sub_of_sub
            complete_occurrences += 1
        else:
            complete_children.append(child)
    if complete_occurrences > 1:
        new_sub_np_lst = []
        optional_lst = []
        for item in sub_np_lst:
            if isinstance(item, list):
                optional_lst.append(item)
            else:
                new_sub_np_lst.append(item)
        sub_np_lst = new_sub_np_lst + optional_lst
    sub_np_lst_couples = []
    set_couple_deps(seq_dict, boundary_np_to_the_left, sub_np_lst_couples, [])
    if sub_np_lst_couples:
        sub_np_lst.append(sub_np_lst_couples)
    for child in complete_children:
        all_sub_of_sub = get_all_valid_sub_np(child, boundary_np_to_the_left)
        sub_np_lst.append(all_sub_of_sub)
    if not sub_np_lst:
        return []
    return sub_np_lst


def get_follow_token(lst_children, token):
    for child in lst_children:
        if child.i == token.i + 1:
            return child
    return None


def get_children_expansion(sub_np_lst, lst_children, boundary_np_to_the_left, head):
    seq_dict = get_seq_deps(lst_children)
    tokens_to_skip = []
    for child in lst_children:
        if child in tokens_to_skip:
            continue
        if child.text in stop_words:
            continue
        is_follow_by_conj_token = False
        if child.i < len(child.doc) - 3 and child.doc[child.i + 1] in lst_children:
            if child.doc[child.i + 1].text in ['/', '-', ',']:
                is_follow_by_conj_token = True
        if child.text.endswith("-") or child.text.endswith("/") or is_follow_by_conj_token:
            follow_token = get_follow_token(lst_children, child)
            if follow_token:
                seq_tokens = [child, follow_token]
                if is_follow_by_conj_token:
                    follow_conj_token = get_follow_token(lst_children, follow_token)
                    if follow_conj_token:
                        seq_tokens.append(follow_conj_token)
                        tokens_to_skip.append(follow_conj_token)
                seq_dict['couple_seq'] = seq_dict.get('couple_seq', [])
                seq_dict['couple_seq'].append(seq_tokens)
                tokens_to_skip.append(follow_token)
                continue
        sub_np = []
        if child.dep_ in dep_type_optional:
            all_sub_of_sub = get_all_valid_sub_np(child, boundary_np_to_the_left)
            sub_np.append(all_sub_of_sub)
            if sub_np:
                sub_np_lst.extend(sub_np)
        elif child.dep_ in combined_with:
            all_sub_of_sub = get_all_valid_sub_special(child, boundary_np_to_the_left)
            if all_sub_of_sub:
                sub_np.append(all_sub_of_sub)
            if sub_np:
                sub_np_lst.extend(sub_np)
        else:
            if child.dep_ not in ['nsubj']:
                continue
    set_couple_deps(seq_dict, boundary_np_to_the_left, sub_np_lst, head)


def get_all_valid_sub_np(head, boundary_np_to_the_left):
    sub_np_lst, lst_children = combine_tied_deps_recursively_and_combine_their_children(head, boundary_np_to_the_left)
    get_children_expansion(sub_np_lst, lst_children, boundary_np_to_the_left, head)
    return sub_np_lst


def get_score(np, head_word):
    val = 0
    for token in np:
        if token == head_word:
            val += 1
            continue
        if token.dep_ in neglect_deps or token.lemma_ in stop_words or token.text == '-':
            continue
        val += 1
    return val


def check_span_validity(np):
    hyphen_lst = []
    idx_lst = []
    for token in np:
        if token.text == '-':
            hyphen_lst.append(token)
            continue
        idx_lst.append(token.i)
    if hyphen_lst:
        for hyphen in hyphen_lst:
            if hyphen.i + 1 in idx_lst and hyphen.i - 1 in idx_lst:
                continue
            else:
                return False
        return True
    return True


def get_all_expansions_of_span_from_lst(span_lst):
    sub_np_final_lst_collection = []
    for head_word, sentence_dep_graph, sentence in span_lst:
        noun_phrase, head_word_in_np_index, boundary_np_to_the_left = valid_expansion_utils.get_np_boundary(
            head_word.i,
            sentence_dep_graph)
        if noun_phrase is None:
            continue
        if len(noun_phrase) > 15:
            continue
        all_valid_sub_np = get_all_valid_sub_np(noun_phrase[head_word_in_np_index],
                                                boundary_np_to_the_left)
        sub_np_final_lst = valid_expansion_utils.from_lst_to_sequence(all_valid_sub_np)
        sub_np_final_spans = []
        for np in sub_np_final_lst:
            np.sort(key=lambda x: x.i)
            is_valid_np = check_span_validity(np)
            if not is_valid_np:
                continue
            val = get_score(np, head_word)
            sub_np_final_spans.append((np, val))
        sub_np_final_spans.sort(key=lambda x: len(x[0]), reverse=True)
        if not sub_np_final_spans:
            continue
        sub_np_final_lst_collection.append((sub_np_final_spans[0][0], head_word, sub_np_final_spans, sentence))
    return sub_np_final_lst_collection



from hierarchybuilder.combine_spans import utils as combine_spans_utils
from hierarchybuilder import utils as ut


def find_similarity_in_same_length_group(lst_spans_tuple):
    black_list = []
    dict_span_to_similar_spans = {}
    for span_tuple in lst_spans_tuple:
        if span_tuple[0] in black_list:
            continue
        dict_span_to_similar_spans[span_tuple[0]] = set()
        dict_span_to_similar_spans[span_tuple[0]].add(span_tuple[0])
        black_list.append(span_tuple[0])
        for span_tuple_to_compare in lst_spans_tuple:
            if span_tuple_to_compare[0] in black_list:
                continue
            is_similar = combine_spans_utils.is_similar_meaning_between_span(span_tuple[1], span_tuple_to_compare[1])
            if is_similar:
                black_list.append(span_tuple_to_compare[0])
                dict_span_to_similar_spans[span_tuple[0]].add(span_tuple_to_compare[0])
    return dict_span_to_similar_spans


def combine_similar_longest_np_with_common_sub_nps(common_np_to_group_members_indices,
                                                   dict_longest_span_to_his_synonyms, dict_span_to_similar_spans):
    black_lst = []
    for span, indices_group in common_np_to_group_members_indices.items():
        if span in black_lst:
            continue
        synonymous_span = dict_span_to_similar_spans[span].intersection(set(dict_longest_span_to_his_synonyms.keys()))
        for longest_span in synonymous_span:
            if longest_span != span and longest_span in common_np_to_group_members_indices.keys():
                common_np_to_group_members_indices[span].update(
                    common_np_to_group_members_indices[longest_span])
                if longest_span not in black_lst:
                    black_lst.append(longest_span)
    for span in black_lst:
        del common_np_to_group_members_indices[span]


def create_dict_from_common_np_to_group_members_indices(span_to_group_members, dict_span_to_rank,
                                                        dict_longest_span_to_his_synonyms, dict_span_to_similar_spans):
    common_np_to_group_members_indices = {k: v for k, v in
                                          sorted(span_to_group_members.items(), key=lambda item: len(item[1]),
                                                 reverse=True)}
    common_np_to_group_members_indices = {k: v for k, v in common_np_to_group_members_indices.items() if
                                          (len(v) > 1 or ut.dict_span_to_counter.get(k, 1) > 1) and
                                          dict_span_to_rank.get(k, 2) >= 2}

    combine_similar_longest_np_with_common_sub_nps(common_np_to_group_members_indices,
                                                   dict_longest_span_to_his_synonyms, dict_span_to_similar_spans)
    return common_np_to_group_members_indices


def create_data_dicts_for_combine_synonyms(label_to_nps_collection, dict_label_to_longest_nps_group):
    span_to_group_members = {}
    dict_longest_span_to_his_synonyms = {}
    for idx, spans_lst in label_to_nps_collection.items():
        for span in spans_lst:
            span_to_group_members[span] = span_to_group_members.get(span, set())
            span_to_group_members[span].add(idx)
        longest_nps_lst = dict_label_to_longest_nps_group[idx]
        for longest_np in longest_nps_lst:
            dict_longest_span_to_his_synonyms[longest_np] = longest_nps_lst
    span_to_group_members = {k: v for k, v in
                             sorted(span_to_group_members.items(), key=lambda item: len(item[1]),
                                    reverse=True)}
    dict_length_to_span = combine_spans_utils.create_dicts_length_to_span_and_span_to_list(span_to_group_members)
    return span_to_group_members, dict_longest_span_to_his_synonyms, dict_length_to_span


def create_dict_span_to_similar_spans_for_all(dict_span_to_similar_spans_partially):
    dict_span_to_similar_spans = {}
    for _, synonyms in dict_span_to_similar_spans_partially.items():
        for span in synonyms:
            dict_span_to_similar_spans[span] = synonyms
    return dict_span_to_similar_spans


def combine_similar_spans(span_to_group_members, dict_length_to_span,
                          dict_longest_span_to_his_synonyms):
    dict_span_to_similar_spans = {}
    for idx, spans in dict_length_to_span.items():
        dict_span_to_similar_spans.update(find_similarity_in_same_length_group(spans))
    span_to_group_members_new = {}
    for span, sub_set in dict_span_to_similar_spans.items():
        span_to_group_members_new[span] = set()
        for synonym in sub_set:
            span_to_group_members_new[span].update(span_to_group_members[synonym])
    for span, synonyms in dict_span_to_similar_spans.items():
        synonyms_intersect = synonyms.intersection(set(dict_longest_span_to_his_synonyms.keys()))
        for synonym in synonyms_intersect:
            synonyms.update(dict_longest_span_to_his_synonyms[synonym])
    dict_span_to_similar_spans = create_dict_span_to_similar_spans_for_all(dict_span_to_similar_spans)
    return span_to_group_members_new, dict_span_to_similar_spans


def union_common_np(label_to_nps_collection, dict_span_to_rank, dict_label_to_longest_nps_group):
    span_to_group_members, dict_longest_span_to_his_synonyms, dict_length_to_span = \
        create_data_dicts_for_combine_synonyms(label_to_nps_collection, dict_label_to_longest_nps_group)
    span_to_group_members, dict_span_to_similar_spans = combine_similar_spans(span_to_group_members,
                                                                              dict_length_to_span,
                                                                              dict_longest_span_to_his_synonyms)
    common_np_to_group_members_indices = \
        create_dict_from_common_np_to_group_members_indices(span_to_group_members,
                                                            dict_span_to_rank, dict_longest_span_to_his_synonyms,
                                                            dict_span_to_similar_spans)
    return common_np_to_group_members_indices, dict_span_to_similar_spans


def union_nps(label_to_nps_collection, dict_span_to_rank, dict_label_to_longest_nps_group):
    common_np_to_group_members_indices, dict_span_to_similar_spans = union_common_np(
        label_to_nps_collection, dict_span_to_rank, dict_label_to_longest_nps_group)
    dict_score_to_collection_of_sub_groups = combine_spans_utils.get_dict_spans_group_to_score(common_np_to_group_members_indices,
                                                                              dict_span_to_rank,
                                                                              dict_span_to_similar_spans,
                                                                              dict_label_to_longest_nps_group)
    return dict_score_to_collection_of_sub_groups, dict_span_to_similar_spans


def create_index_and_collection_for_longest_nps(longest_np_lst, all_nps_example_lst,
                                                global_longest_np_index, global_index_to_similar_longest_np,
                                                longest_NP_to_global_index, dict_uncounted_expansions,
                                                dict_counted_longest_answers):
    label_to_nps_collection = {}
    dict_label_to_longest_nps_group = {}
    if len(longest_np_lst) == 0:
        dict_label_to_longest_nps_group = {}
    elif len(longest_np_lst) == 1:
        longest_NP_to_global_index[longest_np_lst[0]] = global_longest_np_index[0]
        global_index_to_similar_longest_np[global_longest_np_index[0]] = [longest_np_lst[0]]
        dict_label_to_longest_nps_group = {global_longest_np_index[0]:
                                               [longest_np_lst[0]]}
        global_longest_np_index[0] += 1
    else:
        for phrase in all_nps_example_lst:
            longest_span = phrase[0][0]
            all_expansions = []
            for span in phrase:
                all_expansions.append(span[0])
            # longest np clustering
            global_index_to_similar_longest_np[global_longest_np_index[0]] = [longest_span]
            longest_NP_to_global_index[longest_span] = global_longest_np_index[0]
            dict_label_to_longest_nps_group[global_longest_np_index[0]] = [longest_span]
            # Collection of nps
            label_to_nps_collection[global_longest_np_index[0]] = all_expansions
            global_longest_np_index[0] += 1
    if dict_uncounted_expansions:
        label_to_nps_collection.update(dict_uncounted_expansions)
        dict_label_to_longest_nps_group.update(dict_counted_longest_answers)
    return label_to_nps_collection, dict_label_to_longest_nps_group



import torch
from sklearn.cluster import AgglomerativeClustering
from hierarchybuilder.combine_spans import utils as combine_spans_utils

cos = torch.nn.CosineSimilarity(dim=0, eps=1e-08)


def combine_equivalent_nodes(np_object_lst, span_to_object, dict_object_to_global_label, global_dict_label_to_object,
                             topic_object_lst):
    if len(np_object_lst) <= 1:
        return
    node_vector_lst = []
    for np_object in np_object_lst:
        node_vector_lst.append(np_object.weighted_average_vector)
    node_vector_lst = torch.stack(node_vector_lst).reshape(len(np_object_lst), -1)
    clustering = AgglomerativeClustering(distance_threshold=0.1, n_clusters=None, linkage="average",
                                         metric="cosine", compute_full_tree=True).fit(node_vector_lst)
    dict_cluster_to_common_spans_lst = {}
    for idx, label in enumerate(clustering.labels_):
        dict_cluster_to_common_spans_lst[label] = dict_cluster_to_common_spans_lst.get(label, [])
        dict_cluster_to_common_spans_lst[label].append(np_object_lst[idx])
    for label, equivalent_nodes in dict_cluster_to_common_spans_lst.items():
        if len(equivalent_nodes) == 1:
            continue
        combine_spans_utils.combine_nodes_lst(equivalent_nodes, span_to_object, dict_object_to_global_label,
                                              global_dict_label_to_object, topic_object_lst)


def combine_equivalent_node_with_its_equivalent_children(parent, children, span_to_object, dict_object_to_global_label,
                                                         global_dict_label_to_object, topic_object_lst, span_to_vector,
                                                         visited):
    if len(children) == 0:
        return
    equivalents_children = set()
    for child in children:
        res = cos(parent.weighted_average_vector, child.weighted_average_vector)
        if res > 0.95:
            equivalents_children.add(child)
    if equivalents_children:
        combine_spans_utils.combine_node_with_equivalent_children(parent, equivalents_children, span_to_object,
                                                                  dict_object_to_global_label,
                                                                  global_dict_label_to_object, topic_object_lst,
                                                                  span_to_vector, visited)


def combine_equivalent_parent_and_children_nodes_by_semantic_DL_model(np_object_lst, span_to_object,
                                                                      dict_object_to_global_label,
                                                                      global_dict_label_to_object,
                                                                      topic_object_lst, span_to_vector, visited=set()):
    for np_object in np_object_lst:
        if np_object in visited:
            continue
        visited.add(np_object)
        combine_equivalent_parent_and_children_nodes_by_semantic_DL_model(np_object.children, span_to_object,
                                                                          dict_object_to_global_label,
                                                                          global_dict_label_to_object,
                                                                          topic_object_lst, span_to_vector, visited)
        combine_equivalent_node_with_its_equivalent_children(np_object, np_object.children, span_to_object,
                                                             dict_object_to_global_label,
                                                             global_dict_label_to_object, topic_object_lst,
                                                             span_to_vector, visited)


def combine_equivalent_nodes_by_semantic_DL_model(np_object_lst, topic_object_lst, span_to_object,
                                                  dict_object_to_global_label,
                                                  global_dict_label_to_object, span_to_vector, visited=set()):
    for np_object in np_object_lst:
        if np_object in visited:
            continue
        visited.add(np_object)
        combine_equivalent_nodes(np_object.children, span_to_object, dict_object_to_global_label,
                                 global_dict_label_to_object, topic_object_lst)
        combine_equivalent_nodes_by_semantic_DL_model(np_object.children, topic_object_lst, span_to_object,
                                                      dict_object_to_global_label,
                                                      global_dict_label_to_object, span_to_vector, visited)



from hierarchybuilder.combine_spans import combineSpans, paraphrase_detection_SAP_BERT, combine_nodes_UMLS
from hierarchybuilder import utils as ut
import hierarchybuilder.DAG.hierarchical_structure_algorithms as hierarchical_structure_algorithms
import hierarchybuilder.visualization.json_dag_visualization as json_dag_visualization
import hierarchybuilder.DAG.DAG_utils as DAG_utils
from hierarchybuilder.taxonomy import taxonomies_from_UMLS
import sys

sys.setrecursionlimit(10000)


def initialize_spans_data(all_nps_example_lst):
    dict_idx_to_all_valid_expansions = {}
    dict_idx_to_longest_np = {}
    idx = 0
    for phrase in all_nps_example_lst:
        dict_idx_to_all_valid_expansions[idx] = set()
        dict_idx_to_longest_np[idx] = phrase[0][0]
        for span in phrase:
            dict_idx_to_all_valid_expansions[idx].add(span[0])
        idx += 1
    return dict_idx_to_all_valid_expansions, dict_idx_to_longest_np


def get_uncounted_examples(example_list, global_longest_np_lst, dict_global_longest_np_to_all_counted_expansions,
                           longest_NP_to_global_index, global_index_to_similar_longest_np, dict_span_to_lemma_lst,
                           dict_span_to_rank):
    dict_span_to_all_valid_expansions = {}
    longest_np_lst = []
    longest_np_total_lst = []
    all_nps_example_lst = []
    dict_uncounted_expansions = {}
    dict_counted_longest_answers = {}
    duplicate_longest_answers = set()
    for example in example_list:
        longest_np_total_lst.append(example[0])
        if example[0] in duplicate_longest_answers:
            continue
        duplicate_longest_answers.add(example[0])
        if example[0] in global_longest_np_lst:
            label = longest_NP_to_global_index[example[0]]
            dict_counted_longest_answers[label] = set()
            for longest_answer in global_index_to_similar_longest_np[label]:
                dict_counted_longest_answers[label].add(longest_answer)
            uncounted_expansions = set()
            for item in example[1]:
                if item[0] in dict_global_longest_np_to_all_counted_expansions[example[0]]:
                    continue
                dict_span_to_lemma_lst[item[0]] = item[2]
                dict_span_to_rank[item[0]] = len(item[2])
                dict_global_longest_np_to_all_counted_expansions[example[0]].add(item[0])
                uncounted_expansions.add(item[0])
            if uncounted_expansions:
                dict_uncounted_expansions[label] = uncounted_expansions
            continue
        dict_global_longest_np_to_all_counted_expansions[
            example[0]] = dict_global_longest_np_to_all_counted_expansions.get(example[0], set())
        for item in example[1]:
            dict_span_to_lemma_lst[item[0]] = item[2]
            dict_span_to_rank[item[0]] = len(item[2])
            dict_global_longest_np_to_all_counted_expansions[example[0]].add(item[0])
        global_longest_np_lst.add(example[0])
        dict_span_to_all_valid_expansions[example[0]] = [item[0] for item in example[1]]
        longest_np_lst.append(example[0])
        all_nps_example_lst.append(example[1])
    return dict_span_to_all_valid_expansions, longest_np_lst, longest_np_total_lst, all_nps_example_lst, \
           dict_uncounted_expansions, dict_counted_longest_answers


def hierarchy_builder(examples, output_file_name='output', entries_number=50, ignore_words=None, device="", host="127.0.0.1",
                      port=5000):
    ut.initialize_data(examples, host, port, ignore_words, output_file_name, entries_number, device)
    dict_span_to_rank = {}
    dict_span_to_lemma_lst = ut.dict_span_to_lemma_lst
    global_dict_label_to_object = {}
    span_to_object = {}
    topic_object_lst = []
    dict_object_to_global_label = {}
    global_index_to_similar_longest_np = {}
    all_spans = set()
    span_to_vector = {}
    global_longest_np_lst = set()
    dict_global_longest_np_to_all_counted_expansions = {}
    global_longest_np_index = [0]
    longest_NP_to_global_index = {}
    # print("Start the detection of equivalent spans by BOW + building the DAG")
    for topic, examples_list in ut.topics_dict.items():
        topic_synonym_lst = set(ut.dict_noun_lemma_to_synonyms[topic])
        for synonym in topic_synonym_lst:
            dict_span_to_rank[synonym] = 1
        dict_span_to_all_valid_expansions, longest_np_lst, longest_np_total_lst, all_nps_example_lst, \
        dict_uncounted_expansions, dict_counted_longest_answers = \
            get_uncounted_examples(examples_list, global_longest_np_lst,
                                   dict_global_longest_np_to_all_counted_expansions, longest_NP_to_global_index,
                                   global_index_to_similar_longest_np, dict_span_to_lemma_lst, dict_span_to_rank)
        label_to_nps_collection, dict_label_to_longest_nps_group = \
            combineSpans.create_index_and_collection_for_longest_nps(longest_np_lst, all_nps_example_lst,
                                                                     global_longest_np_index,
                                                                     global_index_to_similar_longest_np,
                                                                     longest_NP_to_global_index,
                                                                     dict_uncounted_expansions,
                                                                     dict_counted_longest_answers)
        dict_score_to_collection_of_sub_groups, dict_span_to_similar_spans = \
            combineSpans.union_nps(label_to_nps_collection, dict_span_to_rank, dict_label_to_longest_nps_group)
        DAG_utils.insert_examples_of_topic_to_DAG(dict_score_to_collection_of_sub_groups, topic_synonym_lst,
                                                  dict_span_to_lemma_lst, span_to_object, dict_span_to_similar_spans,
                                                  global_dict_label_to_object, topic_object_lst, longest_np_total_lst,
                                                  longest_np_lst, longest_NP_to_global_index,
                                                  dict_object_to_global_label)
    print("End the detection of equivalent spans by BOW + building the DAG")
    ut.get_all_spans(topic_object_lst, all_spans)
    all_spans = list(all_spans)
    DAG_utils.initialize_all_spans_vectors(all_spans, span_to_vector)
    DAG_utils.initialize_nodes_weighted_average_vector(topic_object_lst, global_index_to_similar_longest_np,
                                                       span_to_vector)
    DAG_utils.update_symmetric_relation_in_DAG(topic_object_lst)
    # print("Start to detect equivalent nodes by DL model")
    paraphrase_detection_SAP_BERT.combine_equivalent_nodes_by_semantic_DL_model(topic_object_lst.copy(),
                                                                                topic_object_lst, span_to_object,
                                                                                dict_object_to_global_label,
                                                                                global_dict_label_to_object,
                                                                                span_to_vector)
    print("End to detect equivalent nodes among children of each node by DL model")
    DAG_utils.update_symmetric_relation_in_DAG(topic_object_lst)
    combine_nodes_UMLS.combine_nodes_by_umls_synonymous_spans(span_to_object, dict_object_to_global_label,
                                                              global_dict_label_to_object, topic_object_lst,
                                                              span_to_vector)
    ut.update_nodes_labels(topic_object_lst)
    DAG_utils.update_symmetric_relation_in_DAG(topic_object_lst)
    DAG_utils.initialize_nodes_weighted_average_vector(topic_object_lst, global_index_to_similar_longest_np,
                                                       span_to_vector)
    # Adding taxonomic nodes
    taxonomic_np_objects = taxonomies_from_UMLS.add_taxonomies_to_DAG_by_UMLS(topic_object_lst, dict_span_to_rank,
                                                                              span_to_object, span_to_vector)
    print("End to detect equivalent nodes and taxonomic relations by UMLS")
    DAG_utils.update_symmetric_relation_in_DAG(topic_object_lst)
    DAG_utils.initialize_nodes_weighted_average_vector(topic_object_lst, global_index_to_similar_longest_np,
                                                       span_to_vector)
    paraphrase_detection_SAP_BERT.combine_equivalent_parent_and_children_nodes_by_semantic_DL_model(
        topic_object_lst.copy(), span_to_object, dict_object_to_global_label, global_dict_label_to_object,
        topic_object_lst, span_to_vector)
    print("End to detect close in meaning children and parent nodes by DL model")
    DAG_utils.update_symmetric_relation_in_DAG(topic_object_lst)
    ut.update_nodes_labels(topic_object_lst)
    # DAG Pruning
    hierarchical_structure_algorithms.DAG_contraction_by_set_cover_algorithm(topic_object_lst,
                                                                             global_dict_label_to_object,
                                                                             global_index_to_similar_longest_np)
    print("Finish DAG contraction")
    DAG_utils.remove_redundant_nodes(topic_object_lst)
    print("Redundant nodes were removed")
    ut.update_nodes_labels(topic_object_lst)
    DAG_utils.initialize_nodes_weighted_average_vector(topic_object_lst, global_index_to_similar_longest_np,
                                                       span_to_vector)
    # Entry-point selection
    top_k_topics, already_counted_labels, all_labels = \
        hierarchical_structure_algorithms.extract_top_k_concept_nodes_greedy_algorithm(entries_number, topic_object_lst,
                                                                                       global_index_to_similar_longest_np)
    print("Finish select top k")
    json_dag_visualization.json_dag_visualization(top_k_topics, global_index_to_similar_longest_np,
                                                  taxonomic_np_objects, topic_object_lst)


# def get_words_as_span(words):
#     span = ""
#     idx = 0
#     for word in words:
#         # if idx == 0 and token.tag_ in ['IN', 'TO']:
#         #     continue
#         if idx != 0 and (word != ',' and word != '.'):
#             span += ' '
#         span += word
#         idx += 1
#     return span
#
#
# def convert_to_input_format(file_name):
#     f = open(file_name, 'r', encoding='utf-8')
#     data = json.load(f)
#     example_lst = []
#     for type in data:
#         example_data = type['sub_matches']['main']
#         sentence_as_lst = example_data['words']
#         offset_first = example_data['captures']['arg1']['first']
#         offset_last = example_data['captures']['arg1']['last']
#         span_as_lst = sentence_as_lst[offset_first: offset_last + 1]
#         example_lst.append((get_words_as_span(sentence_as_lst), get_words_as_span(span_as_lst)))
#     return example_lst
#
#
# example_lst = convert_to_input_format('input_files/input_json_files/obesity.json')
#
#
# hierarchy_builder(examples=example_lst, output_file='results/obesity_debug_mode', entries_number=50,
#                   ignore_words=['cause', 'obesity'])


import hierarchybuilder.utils as ut
import hierarchybuilder.DAG.NounPhraseObject as NounPhrase
import hierarchybuilder.DAG.DAG_utils as DAG_utils
import hierarchybuilder.combine_spans.utils as combine_spans_utils
import requests
from hierarchybuilder.topic_clustering import utils_clustering
import json


def create_dict_RB_to_objects_lst(dict_RB_to_objects, np_object, visited, relation_type='RB'):
    if np_object in visited:
        return
    visited.add(np_object)
    try:
        post_data = json.dumps(list(np_object.span_lst))
        dict_response = requests.post(ut.host_and_port + '/get_broader_terms/',
                                      params={"terms": post_data, "relation_type": relation_type},
                                      timeout=3)
        output = dict_response.json()
        broader_terms = output['broader_terms']
        broader_terms_set = set()
        for term in broader_terms:
            broader_terms_set.update(term)
        for term in broader_terms_set:
            dict_RB_to_objects[term] = dict_RB_to_objects.get(term, set())
            dict_RB_to_objects[term].add(np_object)
    except Exception as e:
        print(e)
    for child in np_object.children:
        create_dict_RB_to_objects_lst(dict_RB_to_objects, child, visited)


def is_parent_in_lst(np_object, object_lst, visited=set()):
    visited.add(np_object)
    intersect_parents = np_object.parents.intersection(object_lst)
    if intersect_parents:
        return True
    for parent in np_object.parents:
        if parent in visited:
            continue
        is_parent_in_list = is_parent_in_lst(parent, object_lst, visited)
        if is_parent_in_list:
            return True
    return False


def initialize_span_to_object_dict(dict_span_to_object, np_object, visited):
    if np_object in visited:
        return
    for span in np_object.span_lst:
        dict_span_to_object[span] = np_object
    visited.add(np_object)
    for child in np_object.children:
        initialize_span_to_object_dict(dict_span_to_object, child, visited)


def update_parents_with_new_labels(node, label_lst, visited=set()):
    for parent in node.parents:
        if parent in visited:
            continue
        visited.add(parent)
        parent.label_lst.update(label_lst)
        update_parents_with_new_labels(parent, label_lst, visited)


def link_np_object_to_RB_related_nodes(np_object, object_lst, added_edges, added_taxonomic_relation,
                                       covered_labels_by_new_topics):
    for np_object_NT in object_lst:
        if np_object_NT == np_object:
            continue
        node_ancestors = set()
        combine_spans_utils.get_all_ancestors(np_object, node_ancestors)
        if np_object_NT in node_ancestors:
            continue
        NT_node_ancestors = set()
        combine_spans_utils.get_all_ancestors(np_object_NT, NT_node_ancestors)
        if np_object in NT_node_ancestors:
            continue
        added_edges.append(np_object_NT)
        added_taxonomic_relation.add(np_object_NT)
        covered_labels_by_new_topics.update(np_object_NT.label_lst)
        np_object.add_children([np_object_NT])
        np_object_NT.parents.add(np_object)
    update_parents_with_new_labels(np_object, np_object.label_lst)


def filter_duplicate_relation(dict_RB_to_objects):
    keys_lst = list(dict_RB_to_objects.keys())
    black_lst = set()
    dict_span_to_equivalent = {}
    for idx, key in enumerate(keys_lst):
        if key in black_lst:
            continue
        for idx_ref in range(idx + 1, len(keys_lst)):
            key_ref = keys_lst[idx_ref]
            if key_ref in black_lst:
                continue
            object_lst = dict_RB_to_objects[key]
            object_lst_ref = dict_RB_to_objects[key_ref]
            if 0.67 * len(object_lst) > len(object_lst_ref):
                break
            intersect_lst = object_lst.intersection(object_lst_ref)
            if len(object_lst) != len(intersect_lst):
                continue
            black_lst.add(key_ref)
            dict_span_to_equivalent[key] = dict_span_to_equivalent.get(key, set())
            dict_span_to_equivalent[key].add(key_ref)

    dict_RB_to_objects = {key: dict_RB_to_objects[key] for key in dict_RB_to_objects if
                          key not in black_lst}
    return dict_RB_to_objects, dict_span_to_equivalent


def initialize_taxonomic_relations(topic_objects):
    dict_RB_to_objects = {}
    # Detect spans with broader term in UMLS
    visited = set()
    for np_object in topic_objects:
        create_dict_RB_to_objects_lst(dict_RB_to_objects, np_object, visited)
    dict_RB_to_objects = {k: v for k, v in
                          sorted(dict_RB_to_objects.items(), key=lambda item: len(item[1]),
                                 reverse=True)}
    # Filter objects that their parents expressed by the new taxonomic relations
    for key, object_lst in dict_RB_to_objects.items():
        remove_lst = set()
        for np_object in object_lst:
            is_parent_in_list = is_parent_in_lst(np_object, object_lst)
            if is_parent_in_list:
                remove_lst.add(np_object)
        for np_object in remove_lst:
            object_lst.remove(np_object)
    dict_RB_to_objects = {k: v for k, v in
                          sorted(dict_RB_to_objects.items(), key=lambda item: len(item[1]),
                                 reverse=True)}
    dict_RB_to_objects = {key: dict_RB_to_objects[key] for key in dict_RB_to_objects if
                          len(dict_RB_to_objects[key]) > 1}
    # Filter duplicate relation
    dict_RB_to_objects, dict_span_to_equivalent = filter_duplicate_relation(dict_RB_to_objects)
    return dict_RB_to_objects, dict_span_to_equivalent


def detect_and_update_existing_object_represent_taxonomic_relation(dict_RB_to_objects,
                                                                   dict_span_to_equivalent, dict_span_to_object):
    entries_already_counted = set()
    dict_RB_exist_objects = {}
    for RB, object_lst in dict_RB_to_objects.items():
        if RB in dict_span_to_object:
            dict_RB_exist_objects[RB] = dict_span_to_object[RB]
            entries_already_counted.add(RB)
            continue
        equivalent_span_lst = dict_span_to_equivalent.get(RB, set())
        for equivalent_span in equivalent_span_lst:
            if equivalent_span in dict_span_to_object:
                entries_already_counted.add(RB)
                dict_RB_exist_objects[RB] = dict_span_to_object[equivalent_span]
                break
    added_edges = []
    added_taxonomic_relation = set()
    covered_labels_by_new_topics = set()
    # Link exist np_object_with the new taxonomic relation to other np objects
    counter = 0
    for RB, np_object in dict_RB_exist_objects.items():
        object_lst = dict_RB_to_objects[RB]
        link_np_object_to_RB_related_nodes(np_object, object_lst, added_edges, added_taxonomic_relation,
                                           covered_labels_by_new_topics)
        counter += 1
    return dict_RB_exist_objects, added_edges, added_taxonomic_relation, covered_labels_by_new_topics


def get_most_descriptive_span(nodes_lst, span_lst):
    span_to_vector = {}
    DAG_utils.initialize_all_spans_vectors(span_lst, span_to_vector)
    max_score = 0
    best_span = ""
    max_represented_vector = None
    for span, represented_vector in span_to_vector.items():
        cos_similarity_val = 0.0
        for node in nodes_lst:
            cos_similarity_val += ut.cos(represented_vector, node.weighted_average_vector)
        if cos_similarity_val >= max_score:
            best_span = span
            max_score = cos_similarity_val
            max_represented_vector = represented_vector
    return best_span, max_represented_vector


# Create and add the new taxonomic relation to the DAG
def create_and_add_new_taxonomic_object_to_DAG(dict_RB_exist_objects, dict_RB_to_objects,
                                               dict_span_to_equivalent, dict_span_to_rank, span_to_vector):
    black_lst = set(dict_RB_exist_objects.keys())
    new_taxonomic_np_objects = set()
    for RB, object_lst in dict_RB_to_objects.items():
        if RB in black_lst:
            continue
        equivalent_span_lst = dict_span_to_equivalent.get(RB, set())
        equivalent_span_lst.add(RB)
        equivalent_span_lst = list(equivalent_span_lst)
        span_tuple_lst = []
        span, represented_vector = get_most_descriptive_span(object_lst, equivalent_span_lst)
        span_to_vector[span] = represented_vector
        span_as_doc = ut.nlp(span)
        lemma_lst = utils_clustering.from_tokens_to_lemmas(span_as_doc)
        span_tuple_lst.append((span, lemma_lst))
        dict_span_to_rank[span] = len(lemma_lst)
        label_lst = set()
        for np_object in object_lst:
            label_lst.update(np_object.label_lst)
        new_np_object = NounPhrase.NP(span_tuple_lst, label_lst)
        new_np_object.weighted_average_vector = represented_vector
        new_taxonomic_np_objects.add(new_np_object)
        new_np_object.add_children(object_lst)
        for np_object in object_lst:
            np_object.parents.add(new_np_object)
    return new_taxonomic_np_objects


def covered_by_taxonomic_relation(new_taxonomic_np_objects, added_edges, added_taxonomic_relation,
                                  covered_labels_by_new_topics):
    # Check the coverage by the new components
    for np_object in new_taxonomic_np_objects:
        added_edges.extend(np_object.children)
        added_taxonomic_relation.update(np_object.children)
        covered_labels_by_new_topics.update(np_object.label_lst)


def initialize_nodes_weighted_average_vector(nodes_lst, global_index_to_similar_longest_np):
    for node in nodes_lst:
        DAG_utils.initialize_node_weighted_vector(node)
        node.frequency = DAG_utils.get_frequency_from_labels_lst(global_index_to_similar_longest_np,
                                                                 node.label_lst)


def add_taxonomies_to_DAG_by_UMLS(topic_objects, dict_span_to_rank, dict_span_to_object, span_to_vector):
    dict_RB_to_objects, dict_span_to_equivalent = initialize_taxonomic_relations(topic_objects)
    # Find exist np objects that represent the taxonomic relation
    # print("End initialization of taxonomic relations")
    dict_RB_exist_objects, added_edges, added_taxonomic_relation, covered_labels_by_new_topics = \
        detect_and_update_existing_object_represent_taxonomic_relation(dict_RB_to_objects, dict_span_to_equivalent,
                                                                       dict_span_to_object)
    ut.isCyclic(topic_objects)
    # print("End detect and update existing object represent taxonomic relation")
    DAG_utils.update_symmetric_relation_in_DAG(topic_objects)
    new_taxonomic_np_objects = create_and_add_new_taxonomic_object_to_DAG(dict_RB_exist_objects,
                                                                          dict_RB_to_objects, dict_span_to_equivalent,
                                                                          dict_span_to_rank, span_to_vector)
    # print("End create and add new taxonomic object to DAG")
    topic_objects.extend(new_taxonomic_np_objects)
    # covered_by_taxonomic_relation(new_taxonomic_np_objects, added_edges, added_taxonomic_relation,
    #                               covered_labels_by_new_topics)
    return new_taxonomic_np_objects



from hierarchybuilder import utils as ut
import nltk
from nltk.corpus import wordnet
from hierarchybuilder.DAG import DAG_utils as DAG_utils


def get_synonyms_by_word(word):
    synonyms = set()
    for syn in wordnet.synsets(word):
        for l in syn.lemmas():
            synonyms.add(l.name())
    return synonyms


def from_words_to_lemma_lst(span):
    lemmas_lst = []
    for word in span:
        lemma = ut.dict_word_to_lemma.get(word, None)
        if lemma is None:
            lemma = word
        lemmas_lst.append(lemma)
    return lemmas_lst


def word_contained_in_list_by_edit_distance(word, lst_words_ref):
    for word_ref in lst_words_ref:
        val = nltk.edit_distance(word, word_ref)
        if val / max(len(word), len(word_ref)) <= 0.25:
            return True, word_ref
    return False, None


def compare_edit_distance_of_synonyms(synonyms, token, lemma_ref):
    close_words = set()
    for synonym in synonyms:
        edit_distance = nltk.edit_distance(synonym, token)
        edit_distance_lemma = nltk.edit_distance(synonym, lemma_ref)
        if edit_distance / max(len(token), len(synonym)) <= 0.25:
            close_words.add((synonym, token))
            continue
        if edit_distance_lemma / max(len(lemma_ref), len(synonym)) <= 0.25:
            close_words.add((synonym, lemma_ref))
            continue
    return list(close_words)


def remove_token_if_in_span(token, span):
    if token in span:
        span.remove(token)
        return True
    else:
        synonyms_lst = set(ut.dict_lemma_to_synonyms.get(token, []))
        for word_lemma in span:
            if set(ut.dict_lemma_to_synonyms.get(word_lemma, [])).intersection(synonyms_lst):
                span.remove(word_lemma)
                return True
    return False


def is_similar_meaning_between_span(span_1_lemma_lst, span_2_lemma_lst):
    not_satisfied = []
    span_1_lemma_lst = span_1_lemma_lst.copy()
    span_2_lemma_lst = span_2_lemma_lst.copy()
    for lemma in span_1_lemma_lst:
        is_exist = remove_token_if_in_span(lemma, span_2_lemma_lst)
        if not is_exist:
            not_satisfied.append(lemma)
    if len(not_satisfied) > 2:
        return False
    if len(span_1_lemma_lst) == 1 and not_satisfied:
        return False
    for lemma in not_satisfied:
        is_exist, lemma_to_remove = word_contained_in_list_by_edit_distance(lemma, span_2_lemma_lst)
        if is_exist:
            span_2_lemma_lst.remove(lemma_to_remove)
            continue
        return False
    return True


def create_dicts_length_to_span_and_span_to_list(span_to_group_members):
    dict_length_to_span = {}
    for span, sub_set in span_to_group_members.items():
        span_as_lst = ut.dict_span_to_lemma_lst[span]
        dict_length_to_span[len(span_as_lst)] = dict_length_to_span.get(len(span_as_lst), [])
        dict_length_to_span[len(span_as_lst)].append((span, span_as_lst))
        ut.dict_span_to_lemma_lst[span] = span_as_lst
    return dict_length_to_span


def get_average_value(spans_lst, dict_span_to_rank):
    average_val = 0
    for span in spans_lst:
        average_val += dict_span_to_rank.get(span, 2)
    average_val = average_val / len(spans_lst)
    return int(average_val * 10)


def create_dict_lemma_word2vec_and_edit_distance():
    words_lst = list(ut.dict_word_to_lemma.keys())
    dict_lemma_to_close_words = {}
    counter = 0
    for word, lemma in ut.dict_word_to_lemma.items():
        dict_lemma_to_close_words[word] = []
        # if word not in vocab_word2vec:
        #     counter += 1
        #     continue
        for word_ref_idx in range(counter + 1, len(words_lst)):
            word_ref = words_lst[word_ref_idx]
            lemma_ref = ut.dict_word_to_lemma[word_ref]
            synonyms = [word, lemma] + ut.dict_lemma_to_synonyms[lemma]
            synonyms = list(set(synonyms))
            if lemma == lemma_ref or lemma_ref in synonyms:
                continue
            dict_lemma_to_close_words[word].extend(compare_edit_distance_of_synonyms(synonyms, word_ref, lemma_ref))
        counter += 1
    return dict_lemma_to_close_words


def get_weighted_average_vector_of_some_vectors_embeddings(spans_embeddings, common_np_lst):
    weighted_average_vector = torch.zeros(spans_embeddings[0].shape)
    for idx, embedding_vector in enumerate(spans_embeddings):
        weighted_average_vector += embedding_vector
    weighted_average_vector /= len(common_np_lst)
    return weighted_average_vector


def get_non_clustered_group_numbers(span_to_group_members, dict_label_to_longest_nps_group):
    all_group_numbers = set(dict_label_to_longest_nps_group.keys())
    already_grouped = set()
    common_span_lst = []
    for common_span, group_numbers in span_to_group_members.items():
        already_grouped.update(group_numbers)
        common_span_lst.append(common_span)
    res_group_numbers = [item for item in all_group_numbers if item not in already_grouped]
    dict_label_to_longest_np_without_common_sub_np = {}
    for num in res_group_numbers:
        dict_label_to_longest_np_without_common_sub_np[num] = dict_label_to_longest_nps_group[num]
    return dict_label_to_longest_np_without_common_sub_np, common_span_lst


def get_most_frequent_span(lst_of_spans):
    most_frequent_span_value = -1
    most_frequent_span = None
    for span in lst_of_spans:
        val = ut.dict_span_to_counter.get(span, 0)
        if val > most_frequent_span_value:
            most_frequent_span_value = val
            most_frequent_span = span
    return most_frequent_span


def convert_dict_label_to_spans_to_most_frequent_span_to_label(dict_label_to_spans_group):
    dict_span_to_label = {}
    for label, tuple_of_spans_lst in dict_label_to_spans_group.items():
        spans_lst = [tuple_of_span[0] for tuple_of_span in tuple_of_spans_lst]
        most_frequent_span = get_most_frequent_span(spans_lst)
        dict_span_to_label[most_frequent_span] = label
    return dict_span_to_label


def update_span_to_group_members_with_longest_answers_dict(span_to_group_members, dict_label_to_longest_nps_group,
                                                           dict_span_to_similar_spans):
    dict_longest_answer_to_label_temp = {}
    for label, spans_lst in dict_label_to_longest_nps_group.items():
        most_frequent_span = get_most_frequent_span(spans_lst)
        is_common_span = False
        if most_frequent_span in span_to_group_members:
            continue
        for span, label_lst in span_to_group_members.items():
            if label in label_lst:
                similar_spans_lst = dict_span_to_similar_spans[span]
                intersection_spans_lst = set(spans_lst).intersection(similar_spans_lst)
                if intersection_spans_lst:
                    dict_span_to_similar_spans[span].update(spans_lst)
                    is_common_span = True
                    break
        if not is_common_span:
            dict_longest_answer_to_label_temp[most_frequent_span] = [label]
            dict_span_to_similar_spans[most_frequent_span] = set(spans_lst)

    span_to_group_members.update(dict_longest_answer_to_label_temp)


def get_dict_spans_group_to_score(span_to_group_members, dict_span_to_rank, dict_span_to_similar_spans,
                                  dict_label_to_longest_nps_group):
    update_span_to_group_members_with_longest_answers_dict(span_to_group_members, dict_label_to_longest_nps_group,
                                                           dict_span_to_similar_spans)
    dict_score_to_collection_of_sub_groups = {}
    for key, group in span_to_group_members.items():
        average_val = get_average_value(dict_span_to_similar_spans[key], dict_span_to_rank)
        dict_score_to_collection_of_sub_groups[average_val] = dict_score_to_collection_of_sub_groups.get(
            average_val, [])
        dict_score_to_collection_of_sub_groups[average_val].append((key, set(group)))
    dict_score_to_collection_of_sub_groups = {k: v for k, v in
                                              sorted(dict_score_to_collection_of_sub_groups.items(),
                                                     key=lambda item: item[0], reverse=True)}
    for score, sub_group_lst in dict_score_to_collection_of_sub_groups.items():
        sub_group_lst.sort(key=lambda tup: len(tup[1]), reverse=True)
    return dict_score_to_collection_of_sub_groups


def get_all_ancestors(node, visited):
    visited.add(node)
    for parent in node.parents:
        if parent in visited:
            continue
        get_all_ancestors(parent, visited)


def update_all_ancestors_node_was_combined(node, label_lst, visited=set()):
    for parent in node.parents:
        if parent in visited:
            continue
        visited.add(parent)
        num_of_labels_before_update = len(parent.label_lst)
        parent.label_lst.update(label_lst)
        num_of_labels_after_update = len(parent.label_lst)
        if num_of_labels_before_update == num_of_labels_after_update:
            continue
        update_all_ancestors_node_was_combined(parent, label_lst, visited)


def get_all_descendents(node, visited):
    visited.add(node)
    for child in node.children:
        if child in visited:
            continue
        get_all_descendents(child, visited)


def has_intersection_between_desc_and_anc(node_1, node_2):
    ancestor_lst = set()
    get_all_ancestors(node_1, ancestor_lst)
    descendent_lst = set()
    get_all_descendents(node_2, descendent_lst)
    if ancestor_lst.intersection(descendent_lst):
        return True


def is_combine_create_circle(node_1, node_2):
    has_intersection = has_intersection_between_desc_and_anc(node_1, node_2)
    if has_intersection:
        return True
    has_intersection = has_intersection_between_desc_and_anc(node_2, node_1)
    if has_intersection:
        return True
    return False


def ambiguous_descendent_in_equivalent_nodes(node_1, node_2):
    children_to_remove_from_node_1 = set()
    children_to_remove_from_node_2 = set()
    for child_1 in node_1.children:
        for child_2 in node_2.children:
            if child_1 in child_2.children:
                children_to_remove_from_node_1.add(child_1)
            elif child_2 in child_1.children:
                children_to_remove_from_node_2.add(child_2)
    for child in children_to_remove_from_node_1:
        node_1.children.remove(child)
        child.parents.remove(node_1)
    for child in children_to_remove_from_node_2:
        node_2.children.remove(child)
        if node_2 in child.parents:
            child.parents.remove(node_2)


def initialize_node_weighted_vector(node, span_to_vector):
    if not node.span_lst:
        return
    is_first = True
    weighted_average_vector = None
    for np in node.span_lst:
        if is_first:
            weighted_average_vector = span_to_vector[np]
            is_first = False
        else:
            weighted_average_vector += span_to_vector[np]
    weighted_average_vector /= len(node.span_lst)
    node.weighted_average_vector = weighted_average_vector


def has_another_path_between_nodes(children, child):
    rest_children = set(children) - {child}
    all_descendents = set()
    has_another_path = False
    for rest_child in rest_children:
        get_all_descendents(rest_child, all_descendents)
        if child in all_descendents:
            has_another_path = True
            break
    return has_another_path


def combine_nodes_lst(np_object_lst, span_to_object, dict_object_to_global_label, global_dict_label_to_object,
                      topic_object_lst, combined_nodes_lst=set()):
    first_element = np_object_lst[0]
    length = len(np_object_lst)
    is_combined = False
    for i in range(1, length):
        second_element = np_object_lst[i]
        if second_element == first_element:
            raise Exception("Parent have duplicate node")
        if second_element in first_element.children:
            if has_another_path_between_nodes(first_element.children, second_element):
                continue
            remove_node_parents_edge(first_element, second_element)
            remove_children_which_already_reached(first_element, second_element)
        elif first_element in second_element.children:
            if has_another_path_between_nodes(second_element.children, first_element):
                continue
            remove_node_parents_edge(second_element, first_element)
            remove_children_which_already_reached(second_element, first_element)
            second_element = first_element
            first_element = np_object_lst[i]
        elif is_combine_create_circle(first_element, second_element):
            continue
        else:
            ambiguous_descendent_in_equivalent_nodes(first_element, second_element)
        combined_nodes_lst.add(second_element)
        first_element.combine_nodes(second_element)
        is_combined = True
        DAG_utils.update_global_dictionaries(span_to_object, global_dict_label_to_object,
                                             dict_object_to_global_label, second_element, first_element)
        if second_element in topic_object_lst:
            topic_object_lst.remove(second_element)
        # for span in second_element.span_lst:
        #     span_to_object[span] = first_element
        # global_label_lst = dict_object_to_global_label.get(hash(second_element), None)
        # if global_label_lst:
        #     dict_object_to_global_label[hash(first_element)] = \
        #         dict_object_to_global_label.get(hash(first_element), set())
        #     for global_label in global_label_lst:
        #         global_dict_label_to_object[global_label] = first_element
        #     dict_object_to_global_label[hash(first_element)].update(global_label_lst)
    if is_combined:
        update_all_ancestors_node_was_combined(first_element, first_element.label_lst)


def update_ans_with_remove_link(node, label_lst, visited=set()):
    if node in visited:
        return
    visited.add(node)
    new_label_lst = node.label_lst - label_lst
    for child in node.children:
        new_label_lst.update(child.label_lst)
    if node.label_lst == new_label_lst:
        return
    node.label_lst = new_label_lst
    for parent in node.parents:
        update_ans_with_remove_link(parent, label_lst, visited)


def remove_node_parents_edge(node, child):
    if node in child.parents:
        child.parents.remove(node)
    while child not in node.children:
        node.children.remove(child)
    # parents_remove_lst = set()
    # for parent in child.parents:
    #     if parent in node.children:
    #         parent.children.remove(child)
    #         parents_remove_lst.add(parent)
    #         update_ans_with_remove_link(parent, child.label_lst)
    # for parent in parents_remove_lst:
    #     child.parents.remove(parent)


def remove_children_which_already_reached(node, combined_node):
    descendents = set()
    get_all_descendents(node, descendents)
    remove_lst = set()
    for child in combined_node.children:
        if child in descendents:
            remove_lst.add(child)
    for child in remove_lst:
        if combined_node in child.parents:
            child.parents.remove(combined_node)
        combined_node.children.remove(child)


def combine_node_with_equivalent_children(node, equivalent_np_object_lst, span_to_object, dict_object_to_global_label,
                                          global_dict_label_to_object, topic_object_lst, span_to_vector,
                                          combined_nodes_lst=set()):
    is_combined = False
    children = set(node.children.copy())
    for child in equivalent_np_object_lst:
        has_another_path = has_another_path_between_nodes(children, child)
        if has_another_path:
            continue
        children.remove(child)
        remove_node_parents_edge(node, child)
        remove_children_which_already_reached(node, child)
        combined_nodes_lst.add(child)
        node.combine_nodes(child)
        is_combined = True
        # for span in child.span_lst:
        #     span_to_object[span] = node
        DAG_utils.update_global_dictionaries(span_to_object, global_dict_label_to_object,
                                             dict_object_to_global_label, child, node)
        if child in topic_object_lst:
            topic_object_lst.remove(child)
        # global_label_lst = dict_object_to_global_label.get(hash(child), None)
        # if global_label_lst:
        #     dict_object_to_global_label[hash(node)] = \
        #         dict_object_to_global_label.get(hash(node), set())
        #     for global_label in global_label_lst:
        #         global_dict_label_to_object[global_label] = node
        #     dict_object_to_global_label[hash(node)].update(global_label_lst)
        # del child
    if is_combined:
        update_all_ancestors_node_was_combined(node, node.label_lst)
        initialize_node_weighted_vector(node, span_to_vector)



from hierarchybuilder.combine_spans import utils as combine_spans_utils
from hierarchybuilder import utils as ut
import torch


def is_span_as_lemma_already_exist(span_as_lemmas, span_as_lemmas_lst):
    for span in span_as_lemmas_lst:
        if len(set(span)) != len(set(span_as_lemmas)):
            continue
        intersection_span = set(span).intersection(set(span_as_lemmas))
        if len(intersection_span) == len(set(span_as_lemmas)):
            return True
    return False


def nps_lst_to_string(nps_lst):
    span_lst = set()
    list_of_span_as_lemmas_lst = []
    for np in nps_lst:
        span_lst.add(np[0])
        if not is_span_as_lemma_already_exist(np[1], list_of_span_as_lemmas_lst):
            list_of_span_as_lemmas_lst.append(np[1])
    return span_lst, list_of_span_as_lemmas_lst


class NP:
    def __init__(self, np, label_lst):
        self.span_lst, self.list_of_span_as_lemmas_lst = nps_lst_to_string(np)
        self.lemma_to_occurrences_dict = {}
        self.dict_key_to_synonyms_keys = {}
        self.calculate_common_denominators_of_spans(self.list_of_span_as_lemmas_lst)
        self.common_lemmas_in_spans = []
        self.length = 0
        self.update_top_lemmas_in_spans()
        self.label_lst = set(label_lst)
        self.children = []
        self.parents = set()
        self.frequency = 0
        self.score = 0.0
        self.marginal_val = 0.0
        self.combined_nodes_lst = set()
        self.weighted_average_vector = torch.zeros(768, 1)

    def get_average_length_of_lemmas_lst(self):
        sum_length_lemma_lst = 0
        for span_as_lemmas_lst in self.list_of_span_as_lemmas_lst:
            sum_length_lemma_lst += len(span_as_lemmas_lst)
        return sum_length_lemma_lst / len(self.list_of_span_as_lemmas_lst)

    def does_lemma_lst_contain_lemma(self, lemma_lst, lemma):
        synonyms = self.dict_key_to_synonyms_keys[lemma]
        if set(lemma_lst).intersection(set(synonyms)):
            return True
        return False

    def get_all_span_as_lemma_lst_contain_common_lemma(self):
        contain_common_lemma_lst = self.list_of_span_as_lemmas_lst.copy()
        for common_lemma in self.common_lemmas_in_spans:
            synonyms = self.dict_key_to_synonyms_keys[common_lemma]
            new_contain_common_lemma_lst = []
            for span_as_lemmas_lst in contain_common_lemma_lst:
                if set(span_as_lemmas_lst).intersection(set(synonyms)):
                    new_contain_common_lemma_lst.append(span_as_lemmas_lst)
                contain_common_lemma_lst = new_contain_common_lemma_lst
        return contain_common_lemma_lst

    def get_best_match(self, average_val):
        contain_common_lemma_lst = self.get_all_span_as_lemma_lst_contain_common_lemma()
        self.lemma_to_occurrences_dict = {k: v for k, v in
                                          sorted(self.lemma_to_occurrences_dict.items(), key=lambda item: item[1],
                                                 reverse=True)}
        for lemma, occurrences in self.lemma_to_occurrences_dict.items():
            is_added = False
            if lemma in self.common_lemmas_in_spans:
                continue
            new_contain_common_lemma_lst = []
            for lemma_lst in contain_common_lemma_lst:
                if self.does_lemma_lst_contain_lemma(lemma_lst, lemma):
                    new_contain_common_lemma_lst.append(lemma_lst)
                    is_added = True
            if is_added and (len(new_contain_common_lemma_lst) > 1 or len(self.common_lemmas_in_spans) == 0):
                self.common_lemmas_in_spans.append(lemma)
                contain_common_lemma_lst = new_contain_common_lemma_lst
            if len(self.common_lemmas_in_spans) < average_val * 0.7:
                break
        self.length = len(self.common_lemmas_in_spans)

    def update_top_lemmas_in_spans(self):
        lemma_to_average_occurrences_dict = {key: value / len(self.list_of_span_as_lemmas_lst)
                                             for key, value in self.lemma_to_occurrences_dict.items()}
        self.common_lemmas_in_spans = [key for key, value in lemma_to_average_occurrences_dict.items() if
                                       value > 0.7]
        self.length = len(self.common_lemmas_in_spans)
        average_val = self.get_average_length_of_lemmas_lst()
        if self.length + 0.5 < average_val:
            self.get_best_match(average_val)
            # most_frequent_span = combine_spans_utils.get_most_frequent_span(self.span_lst)
            # self.common_lemmas_in_spans = ut.dict_span_to_lemma_lst[most_frequent_span]

    def get_lemmas_synonyms_in_keys(self, lemma):
        lemmas_keys_lst = list(self.lemma_to_occurrences_dict.keys())
        for lemma_key in lemmas_keys_lst:
            for key in self.dict_key_to_synonyms_keys[lemma_key]:
                if lemma in ut.dict_lemma_to_synonyms.get(key, []) or \
                        key in ut.dict_lemma_to_synonyms.get(lemma, []):
                    return lemma_key
            is_similar, lemma_ref = \
                combine_spans_utils.word_contained_in_list_by_edit_distance(lemma,
                                                                            self.dict_key_to_synonyms_keys[lemma_key])
            if is_similar:
                return lemma_key
        return None

    def calculate_common_denominators_of_spans(self, span_as_lemmas_lst_to_update):
        for span_as_lst in span_as_lemmas_lst_to_update:
            already_counted = set()
            for lemma in span_as_lst:
                lemma_key = self.get_lemmas_synonyms_in_keys(lemma)
                if lemma_key in already_counted:
                    continue
                if lemma_key:
                    already_counted.add(lemma_key)
                    self.lemma_to_occurrences_dict[lemma_key] += 1
                    self.dict_key_to_synonyms_keys[lemma_key].add(lemma)
                else:
                    already_counted.add(lemma)
                    self.lemma_to_occurrences_dict[lemma] = 1
                    self.dict_key_to_synonyms_keys[lemma] = set()
                    self.dict_key_to_synonyms_keys[lemma].add(lemma)

    def add_children(self, children):
        for child in children:
            if child == self:
                continue
            if child not in self.children:
                self.children.append(child)
        for child in children:
            self.label_lst.update(child.label_lst)

    def add_unique_lst(self, span_as_tokens_lst):
        new_span_lst = []
        for span_as_tokens in span_as_tokens_lst:
            is_already_exist = is_span_as_lemma_already_exist(span_as_tokens, self.list_of_span_as_lemmas_lst)
            # for span in self.list_of_span_as_lemmas_lst:
            #     if len(set(span)) != len(set(span_as_tokens)):
            #         continue
            #     intersection_span = set(span).intersection(set(span_as_tokens))
            #     if len(intersection_span) == len(span_as_tokens):
            #         is_already_exist = True
            #         break
            if not is_already_exist:
                new_span_lst.append(span_as_tokens)
        if new_span_lst:
            self.list_of_span_as_lemmas_lst.extend(new_span_lst)
            self.calculate_common_denominators_of_spans(new_span_lst)
            self.update_top_lemmas_in_spans()

    def update_parents_label(self, np_object, label_lst, visited):
        for parent_object in np_object.parents:
            if parent_object in visited:
                continue
            parent_object.label_lst.update(label_lst)
            visited.add(parent_object)
            self.update_parents_label(parent_object, label_lst, visited)

    def update_children_with_new_parent(self, children, previous_parent):
        for child in children:
            if previous_parent in child.parents:
                child.parents.remove(previous_parent)
            if child == self:
                continue
            child.parents.add(self)

    def update_parents_with_new_node(self, parents, previous_node):
        for parent in parents:
            if previous_node in parent.children:
                parent.children.remove(previous_node)
            if self == parent:
                continue
            if self not in parent.children:
                parent.children.append(self)

    def combine_nodes(self, np_object):
        self.span_lst.update(np_object.span_lst)
        self.add_unique_lst(np_object.list_of_span_as_lemmas_lst)
        # self.np.extend(np_object.np)
        self.label_lst.update(np_object.label_lst)
        self.update_parents_label(np_object, self.label_lst, set())
        self.update_parents_with_new_node(np_object.parents, np_object)
        self.update_children_with_new_parent(np_object.children, np_object)
        self.parents.update(np_object.parents)
        self.add_children(np_object.children)
        del np_object

    def __gt__(self, ob2):
        return self.marginal_val < ob2.marginal_val



import spacy
from hierarchybuilder.expansions import valid_expansion, valid_expansion_utils
import json

nlp = spacy.load("en_ud_model_sm")


def find_sub_list(sub_lst, lst):
    sll = len(sub_lst)
    for ind in (i for i, e in enumerate(lst) if e == sub_lst[0]):
        if lst[ind:ind + sll] == sub_lst:
            return ind, ind + sll - 1
    return -1, -1


def from_tokens_to_string_list(spacy_doc):
    tokens_lst = []
    for token in spacy_doc:
        tokens_lst.append(token.text)
    return tokens_lst


def get_tokens_from_list_by_indices(start_idx_span, end_idx_span, spacy_doc):
    tokens_lst = []
    for token in spacy_doc:
        tokens_lst.append(token)
    return tokens_lst[start_idx_span:end_idx_span + 1]


def find_spacy_span_in_spacy_sentence(span, sentence):
    span_lst = from_tokens_to_string_list(span)
    sent_lst = from_tokens_to_string_list(sentence)
    return find_sub_list(span_lst, sent_lst)


def get_example_from_txt_format(line):
    temp = line.replace("<e1>", "")
    sentence = temp.replace("</e1>", "")
    span = line[line.find("<e1>") + 4: line.find("</e1>")]
    return sentence, span


def get_sentence_ans_span_from_format(sentence, span):
    sent_as_doc = nlp(sentence)
    span_as_doc = nlp(span)
    start_idx_span, end_idx_span = find_spacy_span_in_spacy_sentence(span_as_doc, sent_as_doc)
    if start_idx_span == -1:
        sentence, span, None, None
    span_as_doc = get_tokens_from_list_by_indices(start_idx_span, end_idx_span, sent_as_doc)
    return sentence, span, sent_as_doc, span_as_doc


def get_tokens_as_span(words):
    span = ""
    idx = 0
    for word in words:
        if idx != 0 and (word != ',' and word != '.'):
            span += ' '
        span += word
        idx += 1
    return span


def get_examples_from_input_file(file_name, is_txt_format):
    examples = []
    if is_txt_format:
        with open(file_name, 'r', encoding='utf-8') as f:
            lines = f.readlines()
            for line in lines:
                sentence, span = get_example_from_txt_format(line)
                examples.append((sentence, span))
    else:
        f = open(file_name, 'r', encoding='utf-8')
        data = json.load(f)
        for type in data:
            for mention in type['mentions']:
                span_as_lst = mention['words'][mention['offsets']['first']: mention['offsets']['last'] + 1]
                examples.append((get_tokens_as_span(mention['words']), get_tokens_as_span(span_as_lst)))
        f.close()
    return examples


def get_examples_as_all_optional_answers_format(examples):
    collection_format_examples = []
    for example in examples:
        sentence, span = example
        sentence, span, sent_as_doc, span_as_doc = get_sentence_ans_span_from_format(sentence, span)
        if sent_as_doc is None:
            # print(sentence)
            # print(span)
            continue
        head_of_span = valid_expansion_utils.get_head_of_span(span_as_doc)
        if head_of_span is None:
            continue
        collection_format_examples.append((head_of_span, sent_as_doc, sentence))
    collection_format_examples = valid_expansion.get_all_expansions_of_span_from_lst(collection_format_examples)
    return collection_format_examples


# def get_examples_from_special_format(file_name='input_files/text_files/chest_pain_causes.txt', is_txt_format=False):
#     examples = get_examples_from_input_file(file_name, is_txt_format)
#     collection_format_examples = get_examples_as_all_valid_answers_format(examples)
#     return collection_format_examples

